import java.lang.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;
import org.apache.crimson.tree.*;  //For UserObject() of class ElementEx

/*
   This software may be freely used for academic purpose only. Please
   contact us for profit use. However, when distributed, the original
   source must be clearly stated, and, when the source code is
   distributed, the copiright notice must be retained and any
   alterations in the code must be clearly marked. No warranty is
   given regarding the quality of this software.
*/

class Snode{
    static final boolean DEBUG = false;
    static final short UNDEF = -1;

    //Fields: 
    String nodeName  = null;     
    String nodeValue = null;
    short  nodeType  = UNDEF;
    Snode firstchild;    
    Snode nextsibling;   
    Snode parent;

    Snode(){
	initAll();
    }

    Snode(String s){
	initAll();
        nodeName  = new String(s);
    }

    Snode(Node domNode){
	initAll();
	copyDomNodeInfo(domNode);
    }

    void initAll(){
        nodeName   = null;
	nodeValue  = null;
	nodeType   = UNDEF;
        firstchild = null;
        nextsibling= null;
        parent     = null;
    }

    Snode copy() {
	Snode newNode = new Snode();
        newNode.nodeName   = this.nodeName;
	newNode.nodeValue  = this.nodeValue;
	newNode.nodeType   = this.nodeType;
        newNode.firstchild = null; //To be set explicitly
        newNode.nextsibling= null; //To be set explicitly
        newNode.parent     = null; //To be set explicitly
	return newNode;
    }

    void copyDomNodeInfo(Node domNode) {
	if (UserData.isVirtualAttrNode(domNode)) {
	    if (DEBUG)
	    System.out.println("debug: copy ATTR node: "+domNode.getNodeName());
	    this.setNodeType(Node.ATTRIBUTE_NODE);
	    this.setNodeName(domNode.getNodeName());
	    this.setNodeValue(domNode.getNodeValue());
	}
	else {
	    if (DEBUG)
	    System.out.println("debug: copy ELEM node: "+domNode.getNodeName());
	    this.setNodeType(domNode.getNodeType());
	    this.setNodeName(domNode.getNodeName());
	    this.setNodeValue(domNode.getNodeValue());
	}
	    if (DEBUG)
		System.out.println("@debug: copyDomNodeInfo: "+
				   "("+this.getNodeType()+", "+
				   this.getNodeName()+", "+
				   this.getNodeValue()+") ");
    }

    public void setNodeName(String value){ nodeName = value; }

    public void setNodeValue(String value){ nodeValue = value; }

    public void setNodeType(short type){ nodeType = type; }

    public String getNodeName(){ return nodeName; }

    public String getNodeValue(){
	if (DEBUG) System.out.println("@debug: getNodeValue: nodeValue=["+nodeValue+"]");
	return nodeValue;
    }

    public short getNodeType(){ return nodeType; }

    public Snode getFirstChild(){ return firstchild; }

    public Snode getNextSibling(){ return nextsibling; }

    public Snode getParent(){ return parent; }

    public Snode getRootNode(){
        Snode target = this;

        while (target.parent != null)
            target = target.parent;

        return target;
    }

    public int getNumChildren() {
	Snode parent = this;
	Snode target = parent.firstchild;
	int count = 0;
	while (target.nextsibling != null) {
	    count++; 
	    target = target.nextsibling;
	}
	return count;
    }

    public Snode getRML(){
        Snode target = getRootNode();
        while (target.firstchild != null){
            target = target.firstchild;
            while (target.nextsibling != null)
                target = target.nextsibling;
        }

        return target;
    }

    private Snode addChild(String name){
	Snode child = new Snode(name);
	return addChild(child);
    }

    private Snode addChild(Snode child){
	Snode parent = this;

        if (parent.firstchild == null){
            parent.firstchild = child;
        }
        else{
            Snode prev = parent.firstchild;
            while(prev.nextsibling != null)
                prev = prev.nextsibling;
            prev.nextsibling = child;
        }
	child.parent = parent;

        return child;
    }

    //Node version: 
    public Snode expand(int depth, Node domNode){
	//System.out.println("@debug: expand: enter");
        Snode target;
	String label = domNode.getNodeName();
	Snode child  = new Snode();
	child.copyDomNodeInfo(domNode);

        if (depth == 0){
            target = this.addChild(child);
        }
        else{
            target = parent;
            depth--;
            while (depth > 0){
                target = target.parent;
                depth--;
            }
            target = target.addChild(child);
        }
        return target;
    }

    public void print(){
	printsub(this.getRootNode(), System.out);
    }

    public void print(PrintStream outStream){
	printsub(this.getRootNode(), outStream);
    }

    String toNodeNameStr() {
	Snode node = this;
	if (node.getNodeType()==Node.ATTRIBUTE_NODE)
	    return "@"+node.getNodeName(); 
	else 
	    return node.getNodeName(); 
    }

    String toNodeNameStrFull() {
	Snode node = this;
	String infoString; 
	if (DEBUG)
	    infoString = 
		": ("+node.getNodeType()+", "+
		node.getNodeName()+", "+
		node.getNodeValue()+") ";
	else
	    infoString = "";
	if (node.getNodeType()==Node.ATTRIBUTE_NODE)
	    return "@"+node.getNodeName()+infoString;
	else 
	    return node.getNodeName()+infoString; 
    }

    public String toString() {
	String name;
	Snode node = this;
	if (node.getNodeType()==Node.ATTRIBUTE_NODE)
	    name = "@"+node.getNodeName();
	else 
	    name = node.getNodeName(); 
	return
	    "<"+name+
	    " type="+node.getNodeType()+
	    " value=\""+PrintDomTree.normalizeText(node.getNodeValue())+
	    //" value=\""+node.getNodeValue()+
	    "\">";
    }

    private void printsub(Snode node, PrintStream outStream){

        outStream.print(node.toNodeNameStr());
        Snode nextnode = node.getFirstChild();

	boolean isFirstChild = true;
	boolean isHasChild = (nextnode != null);
	if (isHasChild)
	    outStream.print("[");
        while (nextnode != null){
	    if (!isFirstChild) outStream.print(", ");
            printsub(nextnode, outStream);
            nextnode = nextnode.getNextSibling();
	    isFirstChild = false;
        }
	if (isHasChild)
	    outStream.print("]");
    }

    //XML printing

    String toStringXML() {
	StringBuffer str = new StringBuffer();
	str.append("[toStringXML: not implemented]");
	return str.toString();
    }

    public String printTagStyle(){

	Snode node = this.getRootNode();
	return printTSSub(node);
    }

    private String printTSSub(Snode node){

	String str;

	if (node.nodeName.equals("#text") == true)
	    str = "#text ";
	else if (node.nodeName.equals("#comment") == true)
	    str = "#comment ";
	else if (node.nodeName.startsWith("[") &&
		 node.nodeName.endsWith("]"))
	    str = node.nodeName.substring(1, node.nodeName.length() - 1);
	else if (node.nodeName.startsWith("@") == true)
	    str = node.printAttributeEdge();
	else{
	    boolean tagclose = false;
	    str = "<" + node.nodeName;
	    for (Snode target = node.firstchild; target != null;
		 target = target.nextsibling){
		if (target.nodeName.startsWith("@") == true)
		    str = str + " " + target.printAttributeEdge();
		else{
		    if (tagclose == false){
			str = str + "> " + printTSSub(target);
			tagclose = true;
		    }
		    else
			str = str + printTSSub(target);
		}
	    }
	    if (tagclose == false)
		str = str + " /> ";
	    else
		str = str + " </" + node.nodeName + "> ";
	}

	return str;
    }

    private String printAttributeEdge(){

	String str;
	if (this.firstchild != null){
	    String attrValue = this.firstchild.nodeName;
	    str = this.nodeName.substring(1) + "=\""
		+ attrValue.substring(1, attrValue.length()-1) + "\"";
	}
	else
	    str = this.nodeName.substring(1);
	return str;
    }
}

